<?php

namespace Nest\Http;

use Nest\Application;
use Nyholm\Psr7\ServerRequest;
use Override;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Http\Server\MiddlewareInterface;
use Psr\Http\Server\RequestHandlerInterface;

/**
 * HTTP service manager.
 */
class Http
{
	/**
	 * Current request instance.
	 * @var ServerRequestInterface
	 */
	private ServerRequestInterface $request;

	/**
	 * @param Application $application The application.
	 * @param HttpConfiguration $configuration HTTP service configuration.
	 */
	public function __construct(protected Application $application, protected HttpConfiguration $configuration)
	{
	}

	/**
	 * Get HTTP debug status.
	 * @return bool HTTP debug status.
	 */
	public function getDebug(): bool
	{
		return $this->configuration->getDebug();
	}

	/**
	 * Parse the current server request from global context.
	 * @return ServerRequestInterface Parsed server request.
	 */
	protected function parseRequest(): ServerRequestInterface
	{
		return $this->request = new ServerRequest(
			$_SERVER["REQUEST_METHOD"],
			$_SERVER["REQUEST_URI"],
			getallheaders(),
			fopen("php://input", "r"),
			"1.1",
			$_SERVER,
		);
	}

	/**
	 * Build a chain of request handlers (with middlewares).
	 * @return RequestHandlerInterface
	 */
	protected function buildCallChain(): RequestHandlerInterface
	{
		// Set the final request handler as the current one.
		$currentHandler = $this->configuration->getRequestHandler();

		foreach ($this->configuration->getMiddlewares() as $middleware)
		{ // Add each middleware.
			// The current handler is calling the middleware which will call the previously current one if needed.
			$currentHandler = new class($middleware, $currentHandler) implements RequestHandlerInterface {
				public function __construct(private MiddlewareInterface $middleware, private RequestHandlerInterface $requestHandler)
				{}

				/**
				 * @inheritDoc
				 */
				#[Override] public function handle(ServerRequestInterface $request): ResponseInterface
				{
					return $this->middleware->process($request, $this->requestHandler);
				}
			};
		}

		// Return the built request handler.
		return $currentHandler;
	}

	/**
	 * Start HTTP engine.
	 * @return void
	 */
	public function __invoke(): void
	{
		// Create and initialize request object.
		$this->parseRequest();

		// Get response from handler and send it.
		$response = $this->buildCallChain()->handle($this->request);
		$this->output($response);
	}

	/**
	 * Output a given response.
	 * @param ResponseInterface $response The response to output.
	 * @return void
	 */
	public function output(ResponseInterface $response): void
	{
		// Send status code.
		http_response_code($response->getStatusCode());

		// Send headers.
		foreach ($response->getHeaders() as $name => $values)
		{ // Send each header.
			foreach ($values as $value)
			{ // Send each header value.
				header("$name: $value", false);
			}
		}

		// Send body.
		echo $response->getBody()->getContents();
	}

	/**
	 * Get current request instance, if there is one.
	 * @return ServerRequestInterface|null
	 */
	public function getRequest(): ?ServerRequestInterface
	{
		return $this->request ?? null;
	}
}
